searchData={"items":[{"type":"task","title":"mix pillar.gen.migration","doc":"Migration Generator (very simple)\n\n```\n$bash> mix pillar.gen.migration new_table\n=> Migration generated priv/pillar_migrations/1592084569_new_table.exs\n```","ref":"Mix.Tasks.Pillar.Gen.Migration.html"},{"type":"module","title":"Pillar","doc":"Elixir client for [ClickHouse](https://clickhouse.tech/), a fast open-source\nOnline Analytical Processing (OLAP) database management system.","ref":"Pillar.html"},{"type":"module","title":"Overview - Pillar","doc":"Pillar provides a straightforward way to interact with ClickHouse databases from Elixir,\nsupporting both direct connections and connection pools. It handles query building, \nparameter substitution, and response parsing.\n\nThe library offers the following core features:\n- Direct connection to ClickHouse servers\n- Connection pooling for improved performance\n- Synchronous and asynchronous query execution\n- Structured data insertion\n- Parameter substitution\n- Query result parsing\n- Migration support","ref":"Pillar.html#module-overview"},{"type":"module","title":"Direct Usage - Pillar","doc":"For simple usage with a direct connection:\n\n```elixir\n# Create a connection\nconn = Pillar.Connection.new(\"http://user:password@localhost:8123/database\")\n\n# Execute a query with parameter substitution\nsql = \"SELECT count(*) FROM users WHERE lastname = {lastname}\"\nparams = %{lastname: \"Smith\"}\n\n{:ok, result} = Pillar.query(conn, sql, params)\n# => [%{\"count(*)\" => 347}]\n```","ref":"Pillar.html#module-direct-usage"},{"type":"module","title":"Connection Pool Usage - Pillar","doc":"For production applications, using a connection pool is recommended:\n\n```elixir\ndefmodule MyApp.ClickHouse do\n  use Pillar,\n    connection_strings: [\n      \"http://user:password@host-1:8123/database\",\n      \"http://user:password@host-2:8123/database\"\n    ],\n    name: __MODULE__,\n    pool_size: 15\nend\n\n# Start the connection pool\nMyApp.ClickHouse.start_link()\n\n# Execute queries using the pool\n{:ok, result} = MyApp.ClickHouse.select(\"SELECT * FROM users WHERE id = {id}\", %{id: 123})\n```","ref":"Pillar.html#module-connection-pool-usage"},{"type":"module","title":"Asynchronous Operations - Pillar","doc":"For non-blocking inserts:\n\n```elixir\n# Using direct connection\nPillar.async_insert(conn, \"INSERT INTO events (user_id, event) VALUES ({user_id}, {event})\", %{\n  user_id: 42,\n  event: \"login\"\n})\n\n# Using connection pool\nMyApp.ClickHouse.async_insert(\"INSERT INTO events (user_id, event) VALUES ({user_id}, {event})\", %{\n  user_id: 42,\n  event: \"login\"\n})\n```","ref":"Pillar.html#module-asynchronous-operations"},{"type":"macro","title":"Pillar.__using__/1","doc":"Defines a ClickHouse connection pool module.\n\nThis macro sets up a module that manages a pool of ClickHouse connections.\nThe generated module provides functions to execute queries against the connection pool,\nhandling connection acquisition and release automatically.","ref":"Pillar.html#__using__/1"},{"type":"macro","title":"Options - Pillar.__using__/1","doc":"- `:connection_strings` - Required. A list of ClickHouse connection URLs\n- `:name` - Optional. The name of the pool (defaults to \"PillarPool\")\n- `:pool_size` - Optional. The number of connections to maintain (defaults to 10)\n- `:pool_timeout` - Optional. Timeout for acquiring a connection from the pool (defaults to 5000ms)\n- `:timeout` - Optional. Default query timeout (defaults to 5000ms)","ref":"Pillar.html#__using__/1-options"},{"type":"macro","title":"Generated Functions - Pillar.__using__/1","doc":"The macro generates the following functions in the module:\n\n- `start_link/1` - Starts the connection pool\n- `select/3` - Executes a SELECT query with JSON formatting\n- `query/3` - Executes an arbitrary SQL query\n- `insert/3` - Executes an INSERT statement\n- `insert_to_table/3` - Inserts data into a specified table\n- `async_query/3` - Asynchronously executes a query\n- `async_insert/3` - Asynchronously executes an INSERT statement\n- `async_insert_to_table/3` - Asynchronously inserts data into a specified table","ref":"Pillar.html#__using__/1-generated-functions"},{"type":"macro","title":"Example - Pillar.__using__/1","doc":"```elixir\ndefmodule MyApp.ClickHouse do\n  use Pillar,\n    connection_strings: [\n      \"http://user:password@host-1:8123/database\",\n      \"http://user:password@host-2:8123/database\"\n    ],\n    name: __MODULE__,\n    pool_size: 15,\n    pool_timeout: 10_000,\n    timeout: 30_000\nend\n```","ref":"Pillar.html#__using__/1-example"},{"type":"function","title":"Pillar.async_insert/4","doc":"Async version of `insert/4` that doesn't wait for a response.\n\nThis function sends an INSERT query to ClickHouse but doesn't wait for a response,\nmaking it suitable for fire-and-forget operations where you don't need to confirm\nthe result. Available only when using connection pools through the `use Pillar` macro.","ref":"Pillar.html#async_insert/4"},{"type":"function","title":"Pillar.async_insert_to_table/4","doc":"Async version of `insert_to_table/4` that doesn't wait for a response.\n\nThis function inserts data into a specified table but doesn't wait for a response,\nmaking it suitable for fire-and-forget operations where you don't need to confirm\nthe result. Available only when using connection pools through the `use Pillar` macro.","ref":"Pillar.html#async_insert_to_table/4"},{"type":"function","title":"Pillar.async_query/4","doc":"Async version of `query/4` that doesn't wait for a response.\n\nThis function sends a query to ClickHouse but doesn't wait for a response,\nmaking it suitable for fire-and-forget operations where you don't need to confirm\nthe result. Available only when using connection pools through the `use Pillar` macro.","ref":"Pillar.html#async_query/4"},{"type":"function","title":"Pillar.insert/4","doc":"Executes an INSERT statement against a ClickHouse database.","ref":"Pillar.html#insert/4"},{"type":"function","title":"Parameters - Pillar.insert/4","doc":"- `connection` - A `Pillar.Connection` struct representing the database connection\n- `query` - The SQL query string with optional parameter placeholders in curly braces `{param}`\n- `params` - A map of parameters to be substituted in the query (default: `%{}`)\n- `options` - A map of options to customize the request (default: `%{}`)","ref":"Pillar.html#insert/4-parameters"},{"type":"function","title":"Options - Pillar.insert/4","doc":"- `:timeout` - Request timeout in milliseconds (default: `5000`)\n- Other options are passed to the connection's URL builder","ref":"Pillar.html#insert/4-options"},{"type":"function","title":"Returns - Pillar.insert/4","doc":"- `{:ok, result}` on success\n- `{:error, reason}` on failure","ref":"Pillar.html#insert/4-returns"},{"type":"function","title":"Examples - Pillar.insert/4","doc":"```elixir\nPillar.insert(conn, \"INSERT INTO users (name, email) VALUES ({name}, {email})\", %{\n  name: \"John Doe\",\n  email: \"john@example.com\"\n})\n```","ref":"Pillar.html#insert/4-examples"},{"type":"function","title":"Pillar.insert_to_table/4","doc":"Inserts data into a specified table, automatically generating the INSERT statement.\n\nThis function allows inserting one record (as a map) or multiple records (as a list of maps)\ninto a ClickHouse table without having to manually construct the SQL INSERT statement.","ref":"Pillar.html#insert_to_table/4"},{"type":"function","title":"Parameters - Pillar.insert_to_table/4","doc":"- `connection` - A `Pillar.Connection` struct representing the database connection\n- `table_name` - The name of the table to insert data into\n- `record_or_records` - A map or list of maps representing the data to insert\n- `options` - A map of options to customize the request (default: `%{}`)","ref":"Pillar.html#insert_to_table/4-parameters"},{"type":"function","title":"Options - Pillar.insert_to_table/4","doc":"- `:timeout` - Request timeout in milliseconds (default: `5000`)\n- `:query_options` - Options for the query builder (e.g., `%{format: :json}`)","ref":"Pillar.html#insert_to_table/4-options"},{"type":"function","title":"Returns - Pillar.insert_to_table/4","doc":"- `{:ok, result}` on success\n- `{:error, reason}` on failure","ref":"Pillar.html#insert_to_table/4-returns"},{"type":"function","title":"Examples - Pillar.insert_to_table/4","doc":"Single record:\n```elixir\nPillar.insert_to_table(conn, \"users\", %{\n  name: \"John Doe\",\n  email: \"john@example.com\",\n  created_at: DateTime.utc_now()\n})\n```\n\nMultiple records:\n```elixir\nPillar.insert_to_table(conn, \"users\", [\n  %{name: \"John Doe\", email: \"john@example.com\"},\n  %{name: \"Jane Smith\", email: \"jane@example.com\"}\n])\n```","ref":"Pillar.html#insert_to_table/4-examples"},{"type":"function","title":"Pillar.query/4","doc":"Executes an arbitrary SQL query against a ClickHouse database.\n\nThis function can be used for any type of query (SELECT, CREATE, ALTER, etc.)\nbut doesn't format the response as JSON. For SELECT queries with formatted \nresults, consider using `select/4` instead.","ref":"Pillar.html#query/4"},{"type":"function","title":"Parameters - Pillar.query/4","doc":"- `connection` - A `Pillar.Connection` struct representing the database connection\n- `query` - The SQL query string with optional parameter placeholders in curly braces `{param}`\n- `params` - A map of parameters to be substituted in the query (default: `%{}`)\n- `options` - A map of options to customize the request (default: `%{}`)","ref":"Pillar.html#query/4-parameters"},{"type":"function","title":"Options - Pillar.query/4","doc":"- `:timeout` - Request timeout in milliseconds (default: `5000`)\n- Other options are passed to the connection's URL builder","ref":"Pillar.html#query/4-options"},{"type":"function","title":"Returns - Pillar.query/4","doc":"- `{:ok, result}` on success\n- `{:error, reason}` on failure","ref":"Pillar.html#query/4-returns"},{"type":"function","title":"Examples - Pillar.query/4","doc":"```elixir\nPillar.query(conn, \"CREATE TABLE users (id UInt64, name String, email String) ENGINE = MergeTree() ORDER BY id\")\n\nPillar.query(conn, \"ALTER TABLE users ADD COLUMN created_at DateTime\")\n```","ref":"Pillar.html#query/4-examples"},{"type":"function","title":"Pillar.select/4","doc":"Executes a SELECT query and returns the results in a structured format.\n\nThis function appends 'FORMAT JSON' to the query, which makes ClickHouse return\nthe results in JSON format, which are then parsed into Elixir data structures.","ref":"Pillar.html#select/4"},{"type":"function","title":"Parameters - Pillar.select/4","doc":"- `connection` - A `Pillar.Connection` struct representing the database connection\n- `query` - The SQL query string with optional parameter placeholders in curly braces `{param}`\n- `params` - A map of parameters to be substituted in the query (default: `%{}`)\n- `options` - A map of options to customize the request (default: `%{}`)","ref":"Pillar.html#select/4-parameters"},{"type":"function","title":"Options - Pillar.select/4","doc":"- `:timeout` - Request timeout in milliseconds (default: `5000`)\n- Other options are passed to the connection's URL builder","ref":"Pillar.html#select/4-options"},{"type":"function","title":"Returns - Pillar.select/4","doc":"- `{:ok, result}` on success, where result is a list of maps representing rows\n- `{:error, reason}` on failure","ref":"Pillar.html#select/4-returns"},{"type":"function","title":"Examples - Pillar.select/4","doc":"```elixir\n{:ok, users} = Pillar.select(conn, \n  \"SELECT id, name, email FROM users WHERE signup_date > {date} LIMIT {limit}\", \n  %{date: \"2023-01-01\", limit: 100}\n)\n\n# users is a list of maps like:\n# [\n#   %{\"id\" => 1, \"name\" => \"John Doe\", \"email\" => \"john@example.com\"},\n#   %{\"id\" => 2, \"name\" => \"Jane Smith\", \"email\" => \"jane@example.com\"}\n# ]\n```","ref":"Pillar.html#select/4-examples"},{"type":"module","title":"Pillar.Connection","doc":"Structure with connection config, such as host, port, user, password and other","ref":"Pillar.Connection.html"},{"type":"function","title":"Pillar.Connection.new/2","doc":"Generates Connection from typical connection string:\n\n```\n%Pillar.Connection{} = Pillar.Connection.new(\"https://user:password@localhost:8123/some_database\")\n\n# in this case \"default\" database is used\n%Pillar.Connection{} = Pillar.Connection.new(\"https://localhost:8123\")\n```","ref":"Pillar.Connection.html#new/2"},{"type":"type","title":"Pillar.Connection.t/0","doc":"","ref":"Pillar.Connection.html#t:t/0"},{"type":"function","title":"Pillar.Connection.url_from_connection/2","doc":"","ref":"Pillar.Connection.html#url_from_connection/2"},{"type":"module","title":"Pillar.BulkInsertBuffer","doc":"This module provides functionality for bulk inserts and buffering records\n\n```elixir\ndefmodule BulkToLogs do\n  use Pillar.BulkInsertBuffer,\n    pool: ClickhouseMaster,\n    table_name: \"logs\",\n    interval_between_inserts_in_seconds: 5,\n    on_errors: &__MODULE__.dump_to_file/2\n\n  def dump_to_file(_result, records) do\n    File.write(\"bad_inserts/2025-05-28 21:57:09.918203Z\", inspect(records))\n  end\nend\n```\n\n```elixir\n:ok = BulkToLogs.insert(%{value: \"online\", count: 133, datetime: DateTime.utc_now()})\n:ok = BulkToLogs.insert(%{value: \"online\", count: 134, datetime: DateTime.utc_now()})\n:ok = BulkToLogs.insert(%{value: \"online\", count: 132, datetime: DateTime.utc_now()})\n....\n\n# all this records will be inserted with 5 second interval\n```","ref":"Pillar.BulkInsertBuffer.html"},{"type":"module","title":"Pillar.Migrations.Generator","doc":"Migration generator, used at Mix.Tasks","ref":"Pillar.Migrations.Generator.html"},{"type":"function","title":"Pillar.Migrations.Generator.migration_filepath/2","doc":"","ref":"Pillar.Migrations.Generator.html#migration_filepath/2"},{"type":"function","title":"Pillar.Migrations.Generator.migration_template/1","doc":"","ref":"Pillar.Migrations.Generator.html#migration_template/1"},{"type":"module","title":"Pillar.MigrationsMacro","doc":"Migration's mechanism\n\nFor generation migration files use task `Mix.Tasks.Pillar.Gen.Migration`\n\nFor launching migration define own mix task or release task with code below:\n\n```\nconn = Pillar.Connection.new(connection_string)\nPillar.Migrations.migrate(conn)\n```","ref":"Pillar.MigrationsMacro.html"},{"type":"behaviour","title":"Pillar.HttpClient.Adapter","doc":"A behaviour to be implemented by adapters","ref":"Pillar.HttpClient.Adapter.html"},{"type":"callback","title":"Pillar.HttpClient.Adapter.post/3","doc":"A callback to be implemented by adapters","ref":"Pillar.HttpClient.Adapter.html#c:post/3"},{"type":"module","title":"Pillar.HttpClient.Response","doc":"Wrapper for HTTP response","ref":"Pillar.HttpClient.Response.html"},{"type":"module","title":"Pillar.HttpClient.TransportError","doc":"Wrapper for HTTP transport errors, such as timeout and etc","ref":"Pillar.HttpClient.TransportError.html"},{"type":"type","title":"Pillar.HttpClient.TransportError.t/0","doc":"","ref":"Pillar.HttpClient.TransportError.html#t:t/0"},{"type":"extras","title":"Overview","doc":"# Pillar\n\n[![github.com](https://github.com/balance-platform/pillar/workflows/build/badge.svg?branch=master)](https://github.com/balance-platform/pillar/actions)\n[![hex.pm](https://img.shields.io/badge/docs-hexpm-blue.svg)](https://hexdocs.pm/pillar)\n[![hex.pm](https://img.shields.io/hexpm/v/pillar.svg)](https://hex.pm/packages/pillar)\n[![hex.pm](https://img.shields.io/hexpm/dt/pillar.svg)](https://hex.pm/packages/pillar)\n[![hex.pm](https://img.shields.io/hexpm/l/pillar.svg)](https://hex.pm/packages/pillar)\n[![github.com](https://img.shields.io/github/last-commit/balance-platform/pillar.svg)](https://github.com/balance-platform/pillar/commits/master)\n\nElixir client for [ClickHouse](https://clickhouse.tech/), a fast open-source\nOnline Analytical Processing (OLAP) database management system.","ref":"readme.html"},{"type":"extras","title":"Table of Contents - Overview","doc":"- [Getting Started](#getting-started)\n  - [Installation](#installation)\n  - [Basic Usage](#basic-usage)\n- [Features](#features)\n  - [Direct Usage with connection structure](#direct-usage-with-connection-structure)\n  - [Pool of workers](#pool-of-workers)\n  - [Async insert](#async-insert)\n  - [Buffer for periodical bulk inserts](#buffer-for-periodical-bulk-inserts)\n  - [Migrations](#migrations)\n  - [DateTime Timezones](#timezones)\n  - [Switching between HTTP adapters](#http-adapters)\n- [Configuration](#configuration)\n  - [Connection Options](#connection-options)\n  - [Pool Configuration](#pool-configuration)\n  - [HTTP Adapters](#http-adapters)\n- [Advanced Usage](#advanced-usage)\n  - [Bulk Insert Strategies](#bulk-insert-strategies)\n  - [Custom Type Conversions](#custom-type-conversions)\n- [Troubleshooting](#troubleshooting)\n  - [Common Issues](#common-issues)\n  - [Performance Optimization](#performance-optimization)\n- [Contribution](#contribution)","ref":"readme.html#table-of-contents"},{"type":"extras","title":"Getting Started - Overview","doc":"","ref":"readme.html#getting-started"},{"type":"extras","title":"Installation - Overview","doc":"Add `pillar` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:pillar, \"~> 0.39.0\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Basic Usage - Overview","doc":"Here's a simple example to get started with Pillar:\n\n```elixir\n# Create a direct connection\nconn = Pillar.Connection.new(\"http://user:password@localhost:8123/database\")\n\n# Execute a SELECT query\n{:ok, result} = Pillar.select(conn, \"SELECT * FROM users LIMIT 10\")\n\n# Execute a parameterized query\n{:ok, result} = Pillar.query(conn, \n  \"SELECT * FROM users WHERE created_at > {min_date} LIMIT {limit}\", \n  %{min_date: \"2023-01-01\", limit: 100}\n)\n\n# Insert data\n{:ok, _} = Pillar.insert(conn, \n  \"INSERT INTO events (user_id, event_type, created_at) VALUES ({user_id}, {event_type}, {created_at})\",\n  %{user_id: 123, event_type: \"login\", created_at: DateTime.utc_now()}\n)\n```","ref":"readme.html#basic-usage"},{"type":"extras","title":"Features - Overview","doc":"Pillar offers a comprehensive set of features for working with ClickHouse:","ref":"readme.html#features"},{"type":"extras","title":"Direct Usage with connection structure - Overview","doc":"The most straightforward way to use Pillar is by creating a direct connection to your ClickHouse server:\n\n```elixir\n# Create a connection to your ClickHouse server\nconn = Pillar.Connection.new(\"http://user:password@localhost:8123/database\")\n\n# Parameters are passed in curly braces {} in the SQL query\n# The map structure provides values for these parameters\nsql = \"SELECT count(*) FROM users WHERE lastname = {lastname}\"\n\nparams = %{lastname: \"Smith\"}\n\n{:ok, result} = Pillar.query(conn, sql, params)\n\nresult\n#=> [%{\"count(*)\" => 347}]\n```\n\nYou can also specify additional connection options:\n\n```elixir\nconn = Pillar.Connection.new(\n  \"http://user:password@localhost:8123/database\",\n  %{\n    timeout: 30_000,  # Connection timeout in milliseconds\n    pool_timeout: 5_000,  # Pool timeout in milliseconds\n    request_timeout: 60_000  # Request timeout in milliseconds\n  }\n)\n```","ref":"readme.html#direct-usage-with-connection-structure"},{"type":"extras","title":"Pool of workers - Overview","doc":"For production environments, using a connection pool is highly recommended. This approach provides:\n- Efficient connection management\n- Supervised workers\n- Load balancing across multiple ClickHouse servers\n- Better performance under high load\n\n```elixir\ndefmodule ClickhouseMaster do\n  use Pillar,\n    connection_strings: [\n      \"http://user:password@host-master-1:8123/database\",\n      \"http://user:password@host-master-2:8123/database\"\n    ],\n    name: __MODULE__,\n    pool_size: 15,\n    pool_timeout: 10_000,  # Time to wait for a connection from the pool\n    timeout: 30_000  # Default query timeout\nend\n\n# Start the connection pool as part of your supervision tree\nClickhouseMaster.start_link()\n\n# Execute queries using the pool\n{:ok, result} = ClickhouseMaster.select(\"SELECT * FROM users WHERE age > {min_age}\", %{min_age: 21})\n{:ok, _} = ClickhouseMaster.insert(\"INSERT INTO logs (message) VALUES ({message})\", %{message: \"User logged in\"})\n```\n\nThe pool automatically manages connection acquisition and release, and can handle multiple ClickHouse servers for load balancing and high availability.","ref":"readme.html#pool-of-workers"},{"type":"extras","title":"Async insert - Overview","doc":"Asynchronous inserts are useful for non-blocking operations when you don't need to wait for a response. This is particularly valuable for:\n- Logging events\n- Metrics collection\n- Any high-volume insert operations where immediate confirmation isn't required\n\n```elixir\n# Using a connection pool (recommended approach)\nClickhouseMaster.async_insert(\n  \"INSERT INTO events (user_id, event, timestamp) VALUES ({user_id}, {event}, {timestamp})\",\n  %{\n    user_id: user.id,\n    event: \"password_changed\",\n    timestamp: DateTime.utc_now()\n  }\n) # => :ok\n\n# The request is sent and the function returns immediately without waiting for a response\n```\n\nNote: Async inserts are only available when using a connection pool created with `use Pillar`. If you attempt to use `Pillar.async_insert/4` directly with a connection structure, it will raise an error.","ref":"readme.html#async-insert"},{"type":"extras","title":"Buffer for periodical bulk inserts - Overview","doc":"The bulk insert buffer feature allows you to collect records in memory and insert them in batches at specified intervals. This is highly efficient for:\n- High-frequency event logging\n- Metrics collection\n- Any scenario where you need to insert many small records\n\nThis feature requires a [Pool of workers](#pool-of-workers) to be set up first.\n\n```elixir\ndefmodule BulkToLogs do\n  use Pillar.BulkInsertBuffer,\n    # Reference to your Pillar connection pool\n    pool: ClickhouseMaster,\n    \n    # Target table for inserts\n    table_name: \"logs\",\n    \n    # How often to flush buffered records (seconds)\n    # Default is 5 seconds if not specified\n    interval_between_inserts_in_seconds: 5,\n    \n    # Optional error handler function\n    on_errors: &__MODULE__.dump_to_file/2,\n    \n    # Maximum records to buffer before forcing a flush\n    # Optional, defaults to 10000\n    max_buffer_size: 5000\n\n  @doc \"\"\"\n  Error handler that stores failed inserts into a file\n  \n  Parameters:\n  - result: The error result from ClickHouse\n  - records: The batch of records that failed to insert\n  \"\"\"\n  def dump_to_file(_result, records) do\n    timestamp = DateTime.utc_now() |> DateTime.to_string() |> String.replace(\":\", \"-\")\n    directory = \"bad_inserts\"\n    \n    # Ensure the directory exists\n    File.mkdir_p!(directory)\n    \n    # Write failed records to a file\n    File.write(\"#{directory}/#{timestamp}.log\", inspect(records, pretty: true))\n  end\n\n  @doc \"\"\"\n  Alternative error handler that attempts to retry failed inserts\n  \n  Note: Retrying can be risky in case of persistent errors\n  \"\"\"\n  def retry_insert(_result, records) do\n    # Add a short delay before retrying\n    Process.sleep(1000)\n    __MODULE__.insert(records)\n  end\nend\n```\n\nUsage example:\n\n```elixir\n# Records are buffered in memory until the flush interval\n:ok = BulkToLogs.insert(%{value: \"online\", count: 133, datetime: DateTime.utc_now()})\n:ok = BulkToLogs.insert(%{value: \"online\", count: 134, datetime: DateTime.utc_now()})\n:ok = BulkToLogs.insert(%{value: \"offline\", count: 42, datetime: DateTime.utc_now()})\n\n# All these records will be inserted in a single batch after the configured interval (5 seconds by default)\n```\n\nThe `on_errors` parameter is a callback function that will be invoked if an error occurs during bulk insert. This is useful for:\n- Logging failed inserts\n- Writing failed records to a backup location\n- Implementing custom retry logic\n\nThe callback receives two parameters:\n1. The error result from ClickHouse\n2. The batch of records that failed to insert","ref":"readme.html#buffer-for-periodical-bulk-inserts"},{"type":"extras","title":"Migrations - Overview","doc":"Pillar provides a migrations system to help you manage your ClickHouse database schema changes in a version-controlled manner. This feature is particularly useful for:\n- Creating tables\n- Modifying schema\n- Ensuring consistent database setup across environments\n- Tracking schema changes over time\n\n#### Generating Migrations\n\nMigrations can be generated with the mix task:\n\n```bash\nmix pillar.gen.migration create_events_table\n```\n\nThis creates a new migration file in `priv/pillar_migrations` with a timestamp prefix, for example:\n`priv/pillar_migrations/20250528120000_create_events_table.exs`\n\n#### Basic Migration Structure\n\n```elixir\ndefmodule Pillar.Migrations.CreateEventsTable do\n  def up do\n    \"\"\"\n    CREATE TABLE IF NOT EXISTS events (\n      id UUID,\n      user_id UInt64,\n      event_type String,\n      payload String,\n      created_at DateTime\n    ) ENGINE = MergeTree()\n    ORDER BY (created_at, id)\n    \"\"\"\n  end\n\n  # Optional: Implement a down function for rollbacks\n  def down do\n    \"DROP TABLE IF EXISTS events\"\n  end\nend\n```\n\n#### Multi-Statement Migrations\n\nFor complex scenarios where you need to execute multiple statements in a single migration, return a list of strings:\n\n```elixir\ndefmodule Pillar.Migrations.CreateMultipleTables do\n  def up do\n    # For multi-statement migrations, return a list of strings\n    [\n      \"\"\"\n      CREATE TABLE IF NOT EXISTS events (\n        id UUID,\n        user_id UInt64,\n        event_type String,\n        created_at DateTime\n      ) ENGINE = MergeTree()\n      ORDER BY (created_at, id)\n      \"\"\",\n      \n      \"\"\"\n      CREATE TABLE IF NOT EXISTS event_metrics (\n        date Date,\n        event_type String,\n        count UInt64\n      ) ENGINE = SummingMergeTree(count)\n      ORDER BY (date, event_type)\n      \"\"\"\n    ]\n  end\nend\n```\n\nYou can also dynamically generate migrations:\n\n```elixir\ndefmodule Pillar.Migrations.CreateShardedTables do\n  def up do\n    # Generate 5 sharded tables\n    (0..4) |> Enum.map(fn i ->\n      \"\"\"\n      CREATE TABLE IF NOT EXISTS events_shard_#{i} (\n        id UUID,\n        user_id UInt64,\n        event_type String,\n        created_at DateTime\n      ) ENGINE = MergeTree()\n      ORDER BY (created_at, id)\n      \"\"\"\n    end)\n  end\nend\n```\n\n#### Running Migrations\n\nTo run migrations, create a mix task:\n\n```elixir\ndefmodule Mix.Tasks.MigrateClickhouse do\n  use Mix.Task\n  \n  @shortdoc \"Runs ClickHouse migrations\"\n  \n  def run(args) do\n    # Start any necessary applications\n    Application.ensure_all_started(:pillar)\n    \n    # Parse command line arguments if needed\n    {opts, _, _} = OptionParser.parse(args, strict: [env: :string])\n    env = Keyword.get(opts, :env, \"dev\")\n    \n    # Get connection details from your application config\n    connection_string = Application.get_env(:my_project, String.to_atom(\"clickhouse_#{env}_url\"))\n    conn = Pillar.Connection.new(connection_string)\n    \n    # Run the migrations\n    case Pillar.Migrations.migrate(conn) do\n      :ok -> \n        Mix.shell().info(\"Migrations completed successfully\")\n      {:error, reason} -> \n        Mix.shell().error(\"Migration failed: #{inspect(reason)}\")\n        exit({:shutdown, 1})\n    end\n  end\nend\n```\n\nThen run the migrations with:\n\n```bash\nmix migrate_clickhouse\n# Or with environment specification:\nmix migrate_clickhouse --env=prod\n```\n\n#### Migration Tracking\n\nPillar automatically tracks applied migrations in a special table named `pillar_migrations` in your ClickHouse database. This table is created automatically and contains:\n- The migration version (derived from the timestamp)\n- The migration name\n- When the migration was applied","ref":"readme.html#migrations"},{"type":"extras","title":"Timezones - Overview","doc":"Pillar supports timezone-aware DateTime operations when working with ClickHouse. This is particularly important when:\n- Storing and retrieving DateTime values\n- Performing date/time calculations across different time zones\n- Ensuring consistent timestamp handling\n\nTo enable timezone support:\n\n1. Add the `tzdata` dependency to your project:\n\n```elixir\ndefp deps do\n  [\n    {:pillar, \"~> 0.39.0\"},\n    {:tzdata, \"~> 1.1\"}\n  ]\nend\n```\n\n2. Configure Elixir to use the Tzdata timezone database:\n\n```elixir\n# In your config/config.exs\nconfig :elixir, :time_zone_database, Tzdata.TimeZoneDatabase\n```\n\n3. Use timezone-aware DateTime functions in your code:\n\n```elixir\n# Create a timezone-aware DateTime\ndatetime = DateTime.from_naive!(~N[2023-01-15 10:30:00], \"Europe/Berlin\")\n\n# Insert it into ClickHouse\nPillar.insert(conn, \n  \"INSERT INTO events (id, timestamp) VALUES ({id}, {timestamp})\",\n  %{id: 123, timestamp: datetime}\n)\n```\n\nFor more details on DateTime and timezone handling in Elixir, see the [official documentation](https://hexdocs.pm/elixir/1.12/DateTime.html#module-time-zone-database).","ref":"readme.html#timezones"},{"type":"extras","title":"Configuration - Overview","doc":"","ref":"readme.html#configuration"},{"type":"extras","title":"Connection Options - Overview","doc":"When creating a new connection with `Pillar.Connection.new/2`, you can specify various options (parameters):\n\n```elixir\nPillar.Connection.new(\n  \"http://user:password@localhost:8123/database\",\n    # Params take precedence over URI string\n\n    # Authentication options\n    user: \"default\",             # Override username in URL\n    password: \"secret\",          # Override password in URL\n    \n    # Timeout options\n    timeout: 30_000,             # Connection timeout in milliseconds\n    \n    # ClickHouse specific options\n    database: \"my_database\",     # Override database in URL\n    default_format: \"JSON\",      # Default response format\n    \n    # Query execution options\n    max_execution_time: 60,      # Maximum query execution time in seconds\n    max_memory_usage: 10000000,  # Maximum memory usage for query in bytes\n)\n```","ref":"readme.html#connection-options"},{"type":"extras","title":"Pool Configuration - Overview","doc":"When using a connection pool with `use Pillar`, you can configure the following options:\n\n```elixir\ndefmodule MyApp.ClickHouse do\n  use Pillar,\n    # List of ClickHouse servers for load balancing and high availability\n    connection_strings: [\n      \"http://user:password@clickhouse-1:8123/database\",\n      \"http://user:password@clickhouse-2:8123/database\"\n    ],\n    \n    # Pool name (defaults to module name if not specified)\n    name: __MODULE__,\n    \n    # Number of connections to maintain in the pool (default: 10)\n    pool_size: 15,\n    \n    # Maximum overflow connections (calculated as pool_size * 0.3 by default)\n    max_overflow: 5,\n    \n    # Time to wait when acquiring a connection from the pool in ms (default: 5000)\n    pool_timeout: 10_000,\n    \n    # Default query timeout in ms (default: 5000)\n    timeout: 30_000\nend\n```","ref":"readme.html#pool-configuration"},{"type":"extras","title":"HTTP Adapters - Overview","doc":"Pillar provides multiple HTTP adapters for communicating with ClickHouse:\n\n1. **TeslaMintAdapter** (default): Uses Tesla with Mint HTTP client\n2. **HttpcAdapter**: Uses Erlang's built-in `:httpc` module\n\nIf you encounter issues with the default adapter, you can switch to an alternative:\n\n```elixir\n# In your config/config.exs\nconfig :pillar, Pillar.HttpClient, http_adapter: Pillar.HttpClient.HttpcAdapter\n```\n\nYou can also implement your own HTTP adapter by creating a module that implements the `post/3` function and returns either a `%Pillar.HttpClient.Response{}` or a `%Pillar.HttpClient.TransportError{}` struct:\n\n```elixir\ndefmodule MyApp.CustomHttpAdapter do\n  @behaviour Pillar.HttpClient.Adapter\n\n  @impl true\n  def post(url, body, options) do\n    # Implement your custom HTTP client logic\n    # ...\n    \n    # Return a response struct\n    %Pillar.HttpClient.Response{\n      body: response_body,\n      status: 200,\n      headers: [{\"content-type\", \"application/json\"}]\n    }\n  end\nend\n\n# Configure Pillar to use your custom adapter\nconfig :pillar, Pillar.HttpClient, http_adapter: MyApp.CustomHttpAdapter\n```","ref":"readme.html#http-adapters"},{"type":"extras","title":"Advanced Usage - Overview","doc":"","ref":"readme.html#advanced-usage"},{"type":"extras","title":"Bulk Insert Strategies - Overview","doc":"Pillar provides several strategies for handling bulk inserts:\n\n1. **Direct batch insert**: Insert multiple records in a single query\n   ```elixir\n   records = [\n     %{id: 1, name: \"Alice\", score: 85},\n     %{id: 2, name: \"Bob\", score: 92},\n     %{id: 3, name: \"Charlie\", score: 78}\n   ]\n   \n   Pillar.insert_to_table(conn, \"students\", records)\n   ```\n\n2. **Buffered inserts**: Use `Pillar.BulkInsertBuffer` for timed batch processing\n   ```elixir\n   # Define a buffer module as shown in the Buffer section\n   \n   # Then insert records that will be buffered and flushed periodically\n   StudentMetricsBuffer.insert(%{student_id: 123, metric: \"login\", count: 1})\n   ```\n\n3. **Async inserts**: Use `async_insert` for fire-and-forget operations\n   ```elixir\n   ClickHouseMaster.async_insert_to_table(\"event_logs\", %{\n     event: \"page_view\",\n     user_id: 42,\n     timestamp: DateTime.utc_now()\n   })\n   ```","ref":"readme.html#bulk-insert-strategies"},{"type":"extras","title":"Custom Type Conversions - Overview","doc":"Pillar handles type conversions between Elixir and ClickHouse automatically, but you can extend or customize this behavior:\n\n```elixir\n# Convert a custom Elixir struct to a ClickHouse-compatible format\ndefimpl Pillar.TypeConvert.ToClickHouse, for: MyApp.User do\n  def convert(user) do\n    %{\n      id: user.id,\n      name: user.full_name,\n      email: user.email,\n      created_at: DateTime.to_iso8601(user.inserted_at)\n    }\n  end\nend\n```","ref":"readme.html#custom-type-conversions"},{"type":"extras","title":"Troubleshooting - Overview","doc":"","ref":"readme.html#troubleshooting"},{"type":"extras","title":"Common Issues - Overview","doc":"#### Connection Timeouts\n\nIf you experience connection timeouts, consider:\n\n1. Increasing the timeout values:\n   ```elixir\n   conn = Pillar.Connection.new(url, %{timeout: 30_000})\n   ```\n\n2. Checking network connectivity between your application and ClickHouse\n\n3. Verifying ClickHouse server is running and accepting connections:\n   ```bash\n   curl http://clickhouse-server:8123/ping\n   ```\n\n#### Memory Limitations\n\nFor large queries that consume significant memory:\n\n1. Add query settings to limit memory usage:\n   ```elixir\n   Pillar.query(conn, \"SELECT * FROM huge_table\", %{}, %{\n     max_memory_usage: 10000000000,  # 10GB\n     max_execution_time: 300         # 5 minutes\n   })\n   ```\n\n2. Consider using streaming queries (with FORMAT CSV or TabSeparated) for very large result sets\n\n#### Bulk Insert Failures\n\nIf bulk inserts fail:\n\n1. Check your error handler in the `BulkInsertBuffer` configuration\n2. Verify data types match the table schema\n3. Consider reducing batch sizes or increasing the interval between inserts","ref":"readme.html#common-issues"},{"type":"extras","title":"Performance Optimization - Overview","doc":"1. **Use connection pooling**: Always use a connection pool in production environments\n\n2. **Batch inserts**: Group multiple inserts into a single operation when possible\n\n3. **Use async operations**: For high-volume inserts where immediate confirmation isn't necessary\n\n4. **Query optimization**: Leverage ClickHouse's strengths:\n   - Use proper table engines based on your query patterns\n   - Ensure you have appropriate indices\n   - Filter on columns used in the ORDER BY clause\n\n5. **Connection reuse**: Avoid creating new connections for each query","ref":"readme.html#performance-optimization"},{"type":"extras","title":"Contribution - Overview","doc":"Feel free to make a pull request. All contributions are appreciated!","ref":"readme.html#contribution"},{"type":"extras","title":"Getting Started","doc":"# Getting Started with Pillar\n\nThis guide will help you get up and running with Pillar, the Elixir client for ClickHouse.","ref":"getting_started.html"},{"type":"extras","title":"Installation - Getting Started","doc":"Add `pillar` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:pillar, \"~> 0.39.0\"}\n  ]\nend\n```\n\nThen run:\n\n```bash\nmix deps.get\n```","ref":"getting_started.html#installation"},{"type":"extras","title":"Basic Configuration - Getting Started","doc":"","ref":"getting_started.html#basic-configuration"},{"type":"extras","title":"Direct Connection - Getting Started","doc":"The simplest way to use Pillar is with a direct connection:\n\n```elixir\n# Create a connection to your ClickHouse server\nconn = Pillar.Connection.new(\"http://user:password@localhost:8123/database\")\n```\n\nThe connection URL follows this format:\n```\nhttp://[username:password@]host[:port]/database\n```\n\nWhere:\n- `username` and `password` are your ClickHouse credentials (default: default/empty)\n- `host` is the ClickHouse server hostname or IP\n- `port` is the HTTP port (default: 8123)\n- `database` is the database name","ref":"getting_started.html#direct-connection"},{"type":"extras","title":"Connection Options - Getting Started","doc":"You can provide additional options when creating a connection:\n\n```elixir\nconn = Pillar.Connection.new(\n  \"http://user:password@localhost:8123/database\",\n  %{\n    timeout: 30_000,  # Connection timeout in milliseconds\n    max_execution_time: 60,  # Maximum query execution time in seconds\n    database: \"analytics\"  # Override database in URL\n  }\n)\n```","ref":"getting_started.html#connection-options"},{"type":"extras","title":"Basic Operations - Getting Started","doc":"","ref":"getting_started.html#basic-operations"},{"type":"extras","title":"Running Queries - Getting Started","doc":"```elixir\n# Simple query\n{:ok, result} = Pillar.query(conn, \"SELECT 1\")\n\n# Parameterized query\n{:ok, users} = Pillar.select(\n  conn,\n  \"SELECT * FROM users WHERE age > {min_age} LIMIT {limit}\",\n  %{min_age: 21, limit: 100}\n)\n```","ref":"getting_started.html#running-queries"},{"type":"extras","title":"Inserting Data - Getting Started","doc":"```elixir\n# Insert with parameters\n{:ok, _} = Pillar.insert(\n  conn,\n  \"INSERT INTO events (user_id, event_type, created_at) VALUES ({user_id}, {event_type}, {created_at})\",\n  %{user_id: 123, event_type: \"login\", created_at: DateTime.utc_now()}\n)\n\n# Insert a record using a map\n{:ok, _} = Pillar.insert_to_table(\n  conn,\n  \"users\",\n  %{\n    id: 456,\n    name: \"John Doe\",\n    email: \"john@example.com\",\n    created_at: DateTime.utc_now()\n  }\n)\n\n# Insert multiple records\n{:ok, _} = Pillar.insert_to_table(\n  conn,\n  \"users\",\n  [\n    %{id: 1, name: \"Alice\", email: \"alice@example.com\"},\n    %{id: 2, name: \"Bob\", email: \"bob@example.com\"},\n    %{id: 3, name: \"Charlie\", email: \"charlie@example.com\"}\n  ]\n)\n```","ref":"getting_started.html#inserting-data"},{"type":"extras","title":"Understanding Responses - Getting Started","doc":"Most Pillar functions return one of these response patterns:\n\n```elixir\n{:ok, result} # Success with result data\n{:error, reason} # Error with reason\n```\n\nFor example:\n\n```elixir\ncase Pillar.select(conn, \"SELECT * FROM users LIMIT 10\") do\n  {:ok, users} ->\n    # Do something with the users list\n    Enum.each(users, &IO.inspect/1)\n    \n  {:error, error} ->\n    # Handle the error\n    Logger.error(\"Query failed: #{inspect(error)}\")\nend\n```","ref":"getting_started.html#understanding-responses"},{"type":"extras","title":"Next Steps - Getting Started","doc":"Now that you have a basic understanding of how to use Pillar, you might want to explore:\n\n- [Connection Pool](connection_pool.html) for managing multiple connections\n- [Migrations](migrations.html) for managing your ClickHouse schema\n- [Bulk Insert Strategies](bulk_inserts.html) for efficiently loading data","ref":"getting_started.html#next-steps"},{"type":"extras","title":"Connection Pool","doc":"# Connection Pool\n\nFor production environments, using a connection pool is highly recommended. Pillar makes it easy to create and manage a pool of connections to your ClickHouse servers.","ref":"connection_pool.html"},{"type":"extras","title":"Benefits of Connection Pooling - Connection Pool","doc":"- **Performance**: Reusing existing connections reduces overhead\n- **Resource Management**: Controls the number of concurrent connections\n- **Load Balancing**: Distributes queries across multiple ClickHouse servers\n- **Fault Tolerance**: Handles server failures gracefully\n- **Supervision**: Automatically reconnects failed connections","ref":"connection_pool.html#benefits-of-connection-pooling"},{"type":"extras","title":"Setting Up a Connection Pool - Connection Pool","doc":"To create a connection pool, define a module that uses Pillar:\n\n```elixir\ndefmodule MyApp.ClickHouse do\n  use Pillar,\n    connection_strings: [\n      \"http://user:password@clickhouse-1:8123/database\",\n      \"http://user:password@clickhouse-2:8123/database\"\n    ],\n    name: __MODULE__,\n    pool_size: 15\nend\n```\n\nThen start the connection pool, typically in your application's supervision tree:\n\n```elixir\n# In your application.ex\ndef start(_type, _args) do\n  children = [\n    # Other children...\n    MyApp.ClickHouse\n  ]\n  \n  opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n  Supervisor.start_link(children, opts)\nend\n```","ref":"connection_pool.html#setting-up-a-connection-pool"},{"type":"extras","title":"Configuration Options - Connection Pool","doc":"When setting up a connection pool, you can configure the following options:\n\n| Option | Description | Default |\n|--------|-------------|---------|\n| `connection_strings` | List of ClickHouse server URLs | (required) |\n| `name` | Name of the pool | Module name |\n| `pool_size` | Number of connections to maintain | 10 |\n| `pool_timeout` | Time to wait for a connection from the pool (ms) | 5000 |\n| `timeout` | Default query timeout (ms) | 5000 |\n| `max_overflow` | Maximum overflow connections | ~30% of pool_size |\n\nExample with all options:\n\n```elixir\ndefmodule MyApp.ClickHouse do\n  use Pillar,\n    connection_strings: [\n      \"http://user:password@clickhouse-1:8123/database\",\n      \"http://user:password@clickhouse-2:8123/database\"\n    ],\n    name: __MODULE__,\n    pool_size: 20,\n    pool_timeout: 10_000,\n    timeout: 30_000\nend\n```","ref":"connection_pool.html#configuration-options"},{"type":"extras","title":"Using the Connection Pool - Connection Pool","doc":"The Pillar-generated module provides functions that mirror the core Pillar API:\n\n```elixir\n# Execute a SELECT query\n{:ok, users} = MyApp.ClickHouse.select(\n  \"SELECT * FROM users WHERE age > {min_age}\",\n  %{min_age: 21}\n)\n\n# Execute an INSERT query\n{:ok, _} = MyApp.ClickHouse.insert(\n  \"INSERT INTO events (user_id, event) VALUES ({user_id}, {event})\",\n  %{user_id: 123, event: \"login\"}\n)\n\n# Insert data using a map\n{:ok, _} = MyApp.ClickHouse.insert_to_table(\n  \"users\",\n  %{\n    id: 456,\n    name: \"John Doe\",\n    email: \"john@example.com\"\n  }\n)\n```","ref":"connection_pool.html#using-the-connection-pool"},{"type":"extras","title":"Asynchronous Operations - Connection Pool","doc":"One of the key benefits of using a connection pool is the ability to perform asynchronous operations:\n\n```elixir\n# Async insert\nMyApp.ClickHouse.async_insert(\n  \"INSERT INTO logs (event, timestamp) VALUES ({event}, {timestamp})\",\n  %{event: \"page_view\", timestamp: DateTime.utc_now()}\n)\n\n# Async table insert\nMyApp.ClickHouse.async_insert_to_table(\n  \"logs\",\n  %{\n    event: \"page_view\",\n    user_id: 123,\n    timestamp: DateTime.utc_now()\n  }\n)\n```\n\nAsynchronous operations:\n- Return immediately without waiting for a response\n- Are ideal for logging, metrics, and other non-critical writes\n- Improve application responsiveness\n- Reduce backpressure","ref":"connection_pool.html#asynchronous-operations"},{"type":"extras","title":"High Availability and Load Balancing - Connection Pool","doc":"When providing multiple connection strings, Pillar distributes queries across all available servers:\n\n```elixir\ndefmodule MyApp.ClickHouse do\n  use Pillar,\n    connection_strings: [\n      \"http://user:password@clickhouse-1:8123/database\",\n      \"http://user:password@clickhouse-2:8123/database\",\n      \"http://user:password@clickhouse-3:8123/database\"\n    ],\n    pool_size: 15\nend\n```\n\nThis setup provides:\n- Load balancing across all servers\n- Automatic failover if a server becomes unavailable\n- Better query throughput","ref":"connection_pool.html#high-availability-and-load-balancing"},{"type":"extras","title":"Performance Considerations - Connection Pool","doc":"- **Pool Size**: Should typically match your application's concurrency needs\n- **Overflow**: Provides flexibility during traffic spikes\n- **Timeouts**: Balance between responsiveness and resource utilization\n- **Server Count**: More servers improve throughput but require more connections\n\nFor a high-traffic application, consider:\n\n```elixir\ndefmodule MyApp.ClickHouse do\n  use Pillar,\n    connection_strings: [\n      \"http://user:password@clickhouse-1:8123/database\",\n      \"http://user:password@clickhouse-2:8123/database\",\n      \"http://user:password@clickhouse-3:8123/database\"\n    ],\n    pool_size: 50,  # Larger pool\n    pool_timeout: 5_000,  # Short timeout to fail fast\n    timeout: 60_000  # Longer query timeout for complex queries\nend\n```","ref":"connection_pool.html#performance-considerations"},{"type":"extras","title":"Migrations","doc":"# Migrations\n\nPillar provides a migrations system to help you manage your ClickHouse database schema changes in a version-controlled manner.","ref":"migrations.html"},{"type":"extras","title":"Migration Basics - Migrations","doc":"Migrations are Elixir modules that define schema changes. Each migration has:\n\n- A timestamp prefix for ordering\n- An `up` function that applies changes\n- An optional `down` function for rollbacks\n\nPillar automatically tracks which migrations have been applied in a special `pillar_migrations` table.","ref":"migrations.html#migration-basics"},{"type":"extras","title":"Generating Migrations - Migrations","doc":"Use the built-in mix task to generate new migrations:\n\n```bash\nmix pillar.gen.migration create_users_table\n```\n\nThis creates a file in `priv/pillar_migrations` with a timestamp prefix, for example:\n`priv/pillar_migrations/20250528120000_create_users_table.exs`\n\nThe generated file looks like:\n\n```elixir\ndefmodule Pillar.Migrations.CreateUsersTable do\n  def up do\n    # Your migration SQL goes here\n  end\n\n  def down do\n    # Optional: code to roll back this migration\n  end\nend\n```","ref":"migrations.html#generating-migrations"},{"type":"extras","title":"Writing Migrations - Migrations","doc":"","ref":"migrations.html#writing-migrations"},{"type":"extras","title":"Single Statement Migration - Migrations","doc":"For simple migrations with a single SQL statement:\n\n```elixir\ndefmodule Pillar.Migrations.CreateUsersTable do\n  def up do\n    \"\"\"\n    CREATE TABLE IF NOT EXISTS users (\n      id UInt64,\n      name String,\n      email String,\n      created_at DateTime\n    ) ENGINE = MergeTree()\n    ORDER BY id\n    \"\"\"\n  end\n\n  def down do\n    \"DROP TABLE IF EXISTS users\"\n  end\nend\n```","ref":"migrations.html#single-statement-migration"},{"type":"extras","title":"Multi-Statement Migration - Migrations","doc":"For more complex migrations requiring multiple SQL statements, return a list of strings:\n\n```elixir\ndefmodule Pillar.Migrations.CreateAnalyticsTables do\n  def up do\n    [\n      \"\"\"\n      CREATE TABLE IF NOT EXISTS page_views (\n        user_id UInt64,\n        page_url String,\n        timestamp DateTime\n      ) ENGINE = MergeTree()\n      ORDER BY (timestamp, user_id)\n      \"\"\",\n      \n      \"\"\"\n      CREATE TABLE IF NOT EXISTS user_sessions (\n        session_id String,\n        user_id UInt64,\n        start_time DateTime,\n        duration UInt32\n      ) ENGINE = MergeTree()\n      ORDER BY (start_time, session_id)\n      \"\"\"\n    ]\n  end\n\n  def down do\n    [\n      \"DROP TABLE IF EXISTS user_sessions\",\n      \"DROP TABLE IF EXISTS page_views\"\n    ]\n  end\nend\n```","ref":"migrations.html#multi-statement-migration"},{"type":"extras","title":"Dynamic Migrations - Migrations","doc":"You can also generate migrations dynamically:\n\n```elixir\ndefmodule Pillar.Migrations.CreateShardedTables do\n  def up do\n    Enum.map(0..4, fn shard ->\n      \"\"\"\n      CREATE TABLE IF NOT EXISTS events_shard_#{shard} (\n        id UUID,\n        user_id UInt64,\n        event_type String,\n        created_at DateTime\n      ) ENGINE = MergeTree()\n      ORDER BY (created_at, id)\n      \"\"\"\n    end)\n  end\n\n  def down do\n    Enum.map(0..4, fn shard ->\n      \"DROP TABLE IF EXISTS events_shard_#{shard}\"\n    end)\n  end\nend\n```","ref":"migrations.html#dynamic-migrations"},{"type":"extras","title":"Running Migrations - Migrations","doc":"To run migrations, you'll need to create a mix task:\n\n```elixir\ndefmodule Mix.Tasks.Clickhouse.Migrate do\n  use Mix.Task\n  \n  @shortdoc \"Runs ClickHouse migrations\"\n  \n  def run(args) do\n    # Start necessary applications\n    [:postgrex, :ecto, :pillar]\n    |> Enum.each(&Application.ensure_all_started/1)\n    \n    # Parse command-line arguments\n    {opts, _, _} = OptionParser.parse(args, strict: [env: :string])\n    env = Keyword.get(opts, :env, \"dev\")\n    \n    # Get connection URL from config\n    url_key = String.to_atom(\"clickhouse_#{env}_url\")\n    url = Application.get_env(:my_app, url_key)\n    \n    # Create connection and run migrations\n    conn = Pillar.Connection.new(url)\n    \n    case Pillar.Migrations.migrate(conn) do\n      :ok -> \n        Mix.shell().info(\"Migrations completed successfully\")\n      {:error, reason} -> \n        Mix.shell().error(\"Migration failed: #{inspect(reason)}\")\n        exit({:shutdown, 1})\n    end\n  end\nend\n```\n\nThen run the migrations with:\n\n```bash\nmix clickhouse.migrate\n# Or with environment specification\nmix clickhouse.migrate --env=prod\n```","ref":"migrations.html#running-migrations"},{"type":"extras","title":"Migration Tracking - Migrations","doc":"Pillar automatically tracks applied migrations in a table called `pillar_migrations` in your ClickHouse database.\n\nThis table contains:\n- The migration version (derived from the timestamp)\n- The migration name\n- When the migration was applied\n\nYou can query this table to see which migrations have been applied:\n\n```sql\nSELECT * FROM pillar_migrations ORDER BY version\n```","ref":"migrations.html#migration-tracking"},{"type":"extras","title":"Rollbacks - Migrations","doc":"To implement rollbacks, create another mix task:\n\n```elixir\ndefmodule Mix.Tasks.Clickhouse.Rollback do\n  use Mix.Task\n  \n  @shortdoc \"Rolls back ClickHouse migrations\"\n  \n  def run(args) do\n    # Start necessary applications\n    [:postgrex, :ecto, :pillar]\n    |> Enum.each(&Application.ensure_all_started/1)\n    \n    # Parse command-line arguments\n    {opts, _, _} = OptionParser.parse(args, strict: [env: :string, steps: :integer])\n    env = Keyword.get(opts, :env, \"dev\")\n    steps = Keyword.get(opts, :steps, 1)\n    \n    # Get connection URL from config\n    url_key = String.to_atom(\"clickhouse_#{env}_url\")\n    url = Application.get_env(:my_app, url_key)\n    \n    # Create connection and run migrations\n    conn = Pillar.Connection.new(url)\n    \n    case Pillar.Migrations.rollback(conn, steps) do\n      :ok -> \n        Mix.shell().info(\"Rollback completed successfully\")\n      {:error, reason} -> \n        Mix.shell().error(\"Rollback failed: #{inspect(reason)}\")\n        exit({:shutdown, 1})\n    end\n  end\nend\n```\n\nThen roll back the most recent migration with:\n\n```bash\nmix clickhouse.rollback\n# Or roll back multiple migrations\nmix clickhouse.rollback --steps=3\n```","ref":"migrations.html#rollbacks"},{"type":"extras","title":"Best Practices - Migrations","doc":"1. **Keep migrations immutable** after they've been applied to production\n2. **Test migrations** thoroughly in development and staging environments\n3. **Include both `up` and `down` functions** for all migrations\n4. **Use appropriate ClickHouse engines** based on your query patterns\n5. **Comment complex migrations** to explain the purpose and approach\n6. **Consider data preservation** when altering tables\n7. **Avoid long-running migrations** in production","ref":"migrations.html#best-practices"},{"type":"extras","title":"Bulk Insert Strategies","doc":"# Bulk Insert Strategies\n\nClickHouse excels at processing large volumes of data, and Pillar provides several strategies for efficiently inserting data in bulk.","ref":"bulk_inserts.html"},{"type":"extras","title":"Why Bulk Inserts? - Bulk Insert Strategies","doc":"Bulk inserts offer several advantages:\n\n- **Performance**: Much faster than individual inserts\n- **Network Efficiency**: Fewer round trips to the server\n- **Resource Utilization**: Reduced overhead for both client and server\n- **Scalability**: Better handling of high-volume data streams","ref":"bulk_inserts.html#why-bulk-inserts"},{"type":"extras","title":"Available Strategies - Bulk Insert Strategies","doc":"Pillar offers three main approaches for bulk inserts:\n\n1. **Direct Batch Insert**: Insert multiple records in a single query\n2. **Buffered Inserts**: Collect records and insert them periodically\n3. **Async Inserts**: Non-blocking fire-and-forget operations","ref":"bulk_inserts.html#available-strategies"},{"type":"extras","title":"Direct Batch Insert - Bulk Insert Strategies","doc":"The simplest approach is to batch multiple records into a single insert operation:\n\n```elixir\n# Create a list of records\nrecords = [\n  %{id: 1, name: \"Alice\", score: 85},\n  %{id: 2, name: \"Bob\", score: 92},\n  %{id: 3, name: \"Charlie\", score: 78},\n  # ... potentially hundreds or thousands of records\n]\n\n# Insert them all in a single operation\n{:ok, _} = Pillar.insert_to_table(conn, \"students\", records)\n```\n\nThis approach is ideal when:\n- You already have a complete batch of records\n- You need to ensure all records are inserted successfully\n- You want to handle any errors that might occur","ref":"bulk_inserts.html#direct-batch-insert"},{"type":"extras","title":"Buffered Inserts with BulkInsertBuffer - Bulk Insert Strategies","doc":"For streaming data or high-frequency inserts, Pillar provides the `BulkInsertBuffer` module:\n\n```elixir\ndefmodule MyApp.EventBuffer do\n  use Pillar.BulkInsertBuffer,\n    # Reference to your Pillar connection pool\n    pool: MyApp.ClickHouse,\n    \n    # Target table for inserts\n    table_name: \"events\",\n    \n    # How often to flush buffered records (seconds)\n    interval_between_inserts_in_seconds: 5,\n    \n    # Maximum records to buffer before forcing a flush\n    max_buffer_size: 5000,\n    \n    # Optional error handler\n    on_errors: &__MODULE__.handle_errors/2\n\n  def handle_errors(error_result, failed_records) do\n    # Log the error\n    Logger.error(\"Failed to insert records: #{inspect(error_result)}\")\n    \n    # Save failed records for later processing\n    timestamp = DateTime.utc_now() |> DateTime.to_string() |> String.replace(\":\", \"-\")\n    filepath = \"failed_inserts/#{timestamp}.json\"\n    \n    File.mkdir_p!(\"failed_inserts\")\n    File.write!(filepath, Jason.encode!(failed_records))\n  end\nend\n```\n\nUsage:\n\n```elixir\n# Start the buffer process in your supervision tree\nchildren = [\n  # ...\n  MyApp.EventBuffer\n]\n\n# Insert records - they will be buffered and inserted periodically\n:ok = MyApp.EventBuffer.insert(%{\n  user_id: 123,\n  event_type: \"page_view\",\n  url: \"/products\",\n  timestamp: DateTime.utc_now()\n})\n```\n\nThe buffer will:\n1. Collect records in memory\n2. Insert them as a batch every `interval_between_inserts_in_seconds` seconds\n3. Force a flush if the buffer reaches `max_buffer_size` records\n4. Call the error handler if insertions fail\n\nThis approach is ideal for:\n- High-frequency event tracking\n- Metrics collection\n- Log aggregation\n- Any scenario with many small records","ref":"bulk_inserts.html#buffered-inserts-with-bulkinsertbuffer"},{"type":"extras","title":"Asynchronous Inserts - Bulk Insert Strategies","doc":"For non-critical inserts where you don't need confirmation:\n\n```elixir\n# Using a connection pool\nMyApp.ClickHouse.async_insert(\n  \"INSERT INTO logs (event, timestamp) VALUES ({event}, {timestamp})\",\n  %{event: \"page_view\", timestamp: DateTime.utc_now()}\n)\n\n# Or with insert_to_table\nMyApp.ClickHouse.async_insert_to_table(\n  \"logs\",\n  %{event: \"page_view\", timestamp: DateTime.utc_now()}\n)\n```\n\nAsync inserts:\n- Return immediately without waiting for a response\n- Don't provide error handling\n- Reduce backpressure in high-volume scenarios\n- Are only available when using a connection pool","ref":"bulk_inserts.html#asynchronous-inserts"},{"type":"extras","title":"Choosing the Right Strategy - Bulk Insert Strategies","doc":"| Strategy | Pros | Cons | Best For |\n|----------|------|------|----------|\n| Direct Batch | Simple, reliable | Blocking, requires accumulating records | Scheduled data loads, transactions |\n| Buffered | Efficient for streams, error handling | Memory usage, potential data loss on crash | Event tracking, metrics, logging |\n| Async | Non-blocking, minimal overhead | No error handling or confirmation | Non-critical data, monitoring data |","ref":"bulk_inserts.html#choosing-the-right-strategy"},{"type":"extras","title":"Performance Optimization Tips - Bulk Insert Strategies","doc":"1. **Batch Size**: Experiment to find the optimal batch size (typically 1,000-10,000 records)\n2. **Buffer Interval**: Balance between latency and efficiency (5-30 seconds is common)\n3. **Schema Design**: Ensure tables are optimized for insert performance\n4. **Data Types**: Use appropriate data types and avoid excessive string conversion\n5. **Compression**: Consider using compressed formats for very large inserts\n6. **Monitoring**: Watch memory usage and server load during bulk operations","ref":"bulk_inserts.html#performance-optimization-tips"},{"type":"extras","title":"Example: Handling Large CSV Import - Bulk Insert Strategies","doc":"```elixir\ndefmodule MyApp.CsvImporter do\n  def import_file(filename, chunk_size \\\\ 5000) do\n    filename\n    |> File.stream!()\n    |> CSV.decode(headers: true)\n    |> Stream.chunk_every(chunk_size)\n    |> Stream.each(fn chunk ->\n      records = Enum.map(chunk, &transform_record/1)\n      {:ok, _} = MyApp.ClickHouse.insert_to_table(\"imported_data\", records)\n      IO.puts(\"Imported #{length(records)} records\")\n    end)\n    |> Stream.run()\n  end\n  \n  defp transform_record({:ok, map}) do\n    # Transform CSV row into appropriate format for ClickHouse\n    %{\n      id: String.to_integer(map[\"id\"]),\n      name: map[\"name\"],\n      value: String.to_float(map[\"value\"]),\n      timestamp: parse_timestamp(map[\"timestamp\"])\n    }\n  end\n  \n  defp parse_timestamp(timestamp_string) do\n    # Parse timestamp from string\n    {:ok, datetime, _} = DateTime.from_iso8601(timestamp_string)\n    datetime\n  end\nend\n```\n\nUsage:\n\n```elixir\nMyApp.CsvImporter.import_file(\"large_dataset.csv\")\n```","ref":"bulk_inserts.html#example-handling-large-csv-import"},{"type":"extras","title":"Custom Type Conversions","doc":"# Custom Type Conversions\n\nPillar automatically handles conversions between Elixir data types and ClickHouse data types. However, you can extend or customize this behavior for advanced use cases.","ref":"custom_types.html"},{"type":"extras","title":"Default Type Conversions - Custom Type Conversions","doc":"Pillar handles these type conversions out of the box:\n\n| Elixir Type | ClickHouse Type |\n|-------------|----------------|\n| Integer | Int8, Int16, Int32, Int64, UInt8, UInt16, UInt32, UInt64 |\n| Float | Float32, Float64, Decimal |\n| String | String, FixedString, Enum |\n| Boolean | UInt8 (0/1) |\n| DateTime | DateTime, DateTime64 |\n| Date | Date, Date32 |\n| Map | Object, JSON |\n| List | Array |\n| Tuple | Tuple |\n| UUID | UUID |","ref":"custom_types.html#default-type-conversions"},{"type":"extras","title":"Custom Type Implementations - Custom Type Conversions","doc":"","ref":"custom_types.html#custom-type-implementations"},{"type":"extras","title":"Converting Custom Structs to ClickHouse - Custom Type Conversions","doc":"You can implement the `Pillar.TypeConvert.ToClickHouse` protocol for your custom structs:\n\n```elixir\ndefmodule MyApp.User do\n  defstruct [:id, :name, :email, :metadata, :inserted_at]\nend\n\ndefimpl Pillar.TypeConvert.ToClickHouse, for: MyApp.User do\n  def convert(user) do\n    %{\n      id: user.id,\n      name: user.name,\n      email: user.email,\n      metadata: Jason.encode!(user.metadata),\n      created_at: DateTime.to_iso8601(user.inserted_at)\n    }\n  end\nend\n```\n\nNow you can directly insert `User` structs:\n\n```elixir\nuser = %MyApp.User{\n  id: 123,\n  name: \"John Doe\",\n  email: \"john@example.com\",\n  metadata: %{preferences: %{theme: \"dark\"}},\n  inserted_at: DateTime.utc_now()\n}\n\nPillar.insert_to_table(conn, \"users\", user)\n```","ref":"custom_types.html#converting-custom-structs-to-clickhouse"},{"type":"extras","title":"Custom JSON Conversion - Custom Type Conversions","doc":"For specialized JSON formatting:\n\n```elixir\ndefimpl Pillar.TypeConvert.ToClickHouseJson, for: MyApp.User do\n  def convert(user) do\n    %{\n      \"user_id\" => user.id,\n      \"full_name\" => user.name,\n      \"contact\" => %{\n        \"email\" => user.email\n      },\n      \"preferences\" => user.metadata,\n      \"registration_date\" => DateTime.to_unix(user.inserted_at)\n    }\n  end\nend\n```","ref":"custom_types.html#custom-json-conversion"},{"type":"extras","title":"Custom Types for Query Parameters - Custom Type Conversions","doc":"You can also use custom types in query parameters:\n\n```elixir\ndefmodule MyApp.GeoPoint do\n  defstruct [:latitude, :longitude]\n  \n  def new(lat, lon) do\n    %__MODULE__{latitude: lat, longitude: lon}\n  end\nend\n\ndefimpl Pillar.TypeConvert.ToClickHouse, for: MyApp.GeoPoint do\n  def convert(point) do\n    \"#{point.latitude},#{point.longitude}\"\n  end\nend\n```\n\nUsage:\n\n```elixir\npoint = MyApp.GeoPoint.new(52.5200, 13.4050)\n\nPillar.query(\n  conn,\n  \"SELECT * FROM locations WHERE geoDistance(point, {location}) < 1000\",\n  %{location: point}\n)\n```","ref":"custom_types.html#custom-types-for-query-parameters"},{"type":"extras","title":"Working with ClickHouse Arrays - Custom Type Conversions","doc":"To handle ClickHouse arrays efficiently:\n\n```elixir\ndefmodule MyApp.TaggedItem do\n  defstruct [:id, :name, :tags]\nend\n\ndefimpl Pillar.TypeConvert.ToClickHouse, for: MyApp.TaggedItem do\n  def convert(item) do\n    %{\n      id: item.id,\n      name: item.name,\n      tags: Enum.join(item.tags, \",\")  # Convert Elixir list to ClickHouse Array format\n    }\n  end\nend\n```\n\nQuerying arrays:\n\n```elixir\nPillar.query(\n  conn,\n  \"SELECT * FROM items WHERE hasAny(tags, {search_tags})\",\n  %{search_tags: [\"important\", \"featured\"]}\n)\n```","ref":"custom_types.html#working-with-clickhouse-arrays"},{"type":"extras","title":"DateTime Handling - Custom Type Conversions","doc":"ClickHouse has specific requirements for DateTime values. You can customize the conversion:\n\n```elixir\ndefmodule MyApp.TimeRange do\n  defstruct [:start_time, :end_time]\nend\n\ndefimpl Pillar.TypeConvert.ToClickHouse, for: MyApp.TimeRange do\n  def convert(range) do\n    %{\n      start_time: DateTime.to_iso8601(range.start_time),\n      end_time: DateTime.to_iso8601(range.end_time)\n    }\n  end\nend\n```","ref":"custom_types.html#datetime-handling"},{"type":"extras","title":"Extending Existing Types - Custom Type Conversions","doc":"You can also extend existing implementations:\n\n```elixir\ndefimpl Pillar.TypeConvert.ToClickHouse, for: DateTime do\n  # Override the default implementation\n  def convert(datetime) do\n    # Format with microsecond precision\n    Calendar.strftime(datetime, \"%Y-%m-%d %H:%M:%S.%6f\")\n  end\nend\n```","ref":"custom_types.html#extending-existing-types"},{"type":"extras","title":"Custom Decoding of ClickHouse Values - Custom Type Conversions","doc":"To customize how ClickHouse values are converted to Elixir:\n\n```elixir\ndefmodule MyApp.ClickHouseJson do\n  @behaviour Pillar.TypeConvert.ToElixir\n\n  def convert(\"DateTime\", value) do\n    # Custom DateTime parsing\n    {:ok, datetime, _} = DateTime.from_iso8601(value <> \"Z\")\n    datetime\n  end\n  \n  def convert(\"Array(String)\", value) do\n    # Custom array parsing\n    String.split(value, \",\") |> Enum.map(&String.trim/1)\n  end\n  \n  # Fall back to default implementation for other types\n  def convert(type, value) do\n    Pillar.TypeConvert.ToElixir.convert(type, value)\n  end\nend\n\n# Configure Pillar to use your custom converter\nconfig :pillar, :type_converter_to_elixir, MyApp.ClickHouseJson\n```","ref":"custom_types.html#custom-decoding-of-clickhouse-values"},{"type":"extras","title":"Best Practices - Custom Type Conversions","doc":"1. **Keep conversions pure**: Avoid side effects in conversion functions\n2. **Handle errors gracefully**: Consider what happens with invalid data\n3. **Respect ClickHouse types**: Ensure your conversions match the expected format\n4. **Test conversions**: Verify both directions (Elixir to ClickHouse and back)\n5. **Consider performance**: Conversions run for every record, so keep them efficient","ref":"custom_types.html#best-practices"},{"type":"extras","title":"HTTP Adapters","doc":"# HTTP Adapters\n\nPillar supports multiple HTTP adapters for communicating with ClickHouse servers. This allows you to choose the best implementation for your specific requirements or environment.","ref":"http_adapters.html"},{"type":"extras","title":"Available Adapters - HTTP Adapters","doc":"Pillar comes with two built-in HTTP adapters:\n\n1. **Tesla Mint Adapter** (default) - Uses [Tesla](https://github.com/teamon/tesla) with the [Mint](https://github.com/elixir-mint/mint) HTTP client\n2. **Httpc Adapter** - Uses Erlang's built-in `:httpc` module","ref":"http_adapters.html#available-adapters"},{"type":"extras","title":"Adapter Architecture - HTTP Adapters","doc":"Each adapter implements the `Pillar.HttpClient.Adapter` behaviour, which requires a `post/3` function:\n\n```elixir\n@callback post(url :: String.t(), post_body :: String.t(), options :: keyword()) ::\n            Pillar.HttpClient.Response.t()\n            | Pillar.HttpClient.TransportError.t()\n            | %{__struct__: RuntimeError, message: String.t()}\n```","ref":"http_adapters.html#adapter-architecture"},{"type":"extras","title":"Troubleshooting","doc":"# Troubleshooting\n\nThis guide covers common issues you may encounter when using Pillar with ClickHouse, along with solutions and performance optimization tips.","ref":"troubleshooting.html"},{"type":"extras","title":"Connection Issues - Troubleshooting","doc":"","ref":"troubleshooting.html#connection-issues"},{"type":"extras","title":"Failed to Connect to ClickHouse Server - Troubleshooting","doc":"**Problem:** Cannot establish a connection to the ClickHouse server.\n\n**Symptoms:**\n```\n{:error, %Pillar.HttpClient.TransportError{reason: :econnrefused}}\n```\n\n**Solutions:**\n\n1. **Check server availability:**\n   ```bash\n   curl http://your-clickhouse-host:8123/ping\n   ```\n   If this doesn't return \"OK\", the server might be down or unreachable.\n\n2. **Verify credentials:**\n   Ensure your username and password are correct in the connection string.\n\n3. **Check network connectivity:**\n   - Verify your application can reach the ClickHouse host\n   - Check for firewall rules blocking port 8123\n   - Ensure the ClickHouse server is configured to accept external connections\n\n4. **Inspect server logs:**\n   Look at ClickHouse server logs for any errors related to connections or authentication.","ref":"troubleshooting.html#failed-to-connect-to-clickhouse-server"},{"type":"extras","title":"Connection Timeout - Troubleshooting","doc":"**Problem:** Connection attempts to ClickHouse time out.\n\n**Symptoms:**\n```\n{:error, %Pillar.HttpClient.TransportError{reason: :timeout}}\n```\n\n**Solutions:**\n\n1. **Increase timeout value:**\n   ```elixir\n   conn = Pillar.Connection.new(\n     \"http://user:password@localhost:8123/database\",\n     %{timeout: 30_000} # 30 seconds\n   )\n   ```\n\n2. **Check server load:**\n   High server load can cause timeouts. Check ClickHouse's system metrics:\n   ```sql\n   SELECT * FROM system.metrics\n   ```\n\n3. **Network latency:**\n   Consider using a server in the same region/datacenter as your application to reduce latency.","ref":"troubleshooting.html#connection-timeout"},{"type":"extras","title":"Query Execution Issues - Troubleshooting","doc":"","ref":"troubleshooting.html#query-execution-issues"},{"type":"extras","title":"Invalid Query Syntax - Troubleshooting","doc":"**Problem:** SQL syntax errors in queries.\n\n**Symptoms:**\n```\n{:error, \"Code: 62. DB::Exception: Syntax error: failed at position 10 (...)\"\n```\n\n**Solutions:**\n\n1. **Validate SQL syntax:**\n   Test your query directly against ClickHouse using the HTTP interface or clickhouse-client.\n\n2. **Check parameter placeholders:**\n   Ensure all parameter placeholders in the query have corresponding values in the params map:\n   ```elixir\n   # Correct\n   Pillar.query(conn, \"SELECT * FROM users WHERE id = {id}\", %{id: 123})\n   \n   # Incorrect - missing parameter\n   Pillar.query(conn, \"SELECT * FROM users WHERE id = {id}\", %{})\n   ```\n\n3. **ClickHouse SQL specifics:**\n   Remember that ClickHouse SQL dialect has some differences from standard SQL.","ref":"troubleshooting.html#invalid-query-syntax"},{"type":"extras","title":"Out of Memory Errors - Troubleshooting","doc":"**Problem:** Query fails due to insufficient memory.\n\n**Symptoms:**\n```\n{:error, \"Code: 241. DB::Exception: Memory limit (for query) exceeded: ...\"\n```\n\n**Solutions:**\n\n1. **Limit query resources:**\n   ```elixir\n   Pillar.query(conn, \"SELECT * FROM large_table\", %{}, %{\n     max_memory_usage: 10000000000, # 10GB\n     max_bytes_before_external_sort: 5000000000 # 5GB\n   })\n   ```\n\n2. **Add query limits:**\n   ```elixir\n   Pillar.select(conn, \"SELECT * FROM huge_table LIMIT 1000\", %{})\n   ```\n\n3. **Use sampling:**\n   ```elixir\n   Pillar.select(conn, \"SELECT * FROM huge_table SAMPLE 0.1\", %{})\n   ```\n\n4. **Optimize schema:**\n   Ensure your tables are using appropriate engines and sorting keys.","ref":"troubleshooting.html#out-of-memory-errors"},{"type":"extras","title":"Data Insertion Issues - Troubleshooting","doc":"","ref":"troubleshooting.html#data-insertion-issues"},{"type":"extras","title":"Bulk Insert Failures - Troubleshooting","doc":"**Problem:** Bulk insert operations fail with data type errors.\n\n**Symptoms:**\n```\n{:error, \"Code: 53. DB::Exception: Cannot parse input: expected )...\"\n```\n\n**Solutions:**\n\n1. **Validate data types:**\n   Ensure all values match the expected column types in your table.\n\n2. **Check schema compatibility:**\n   ```elixir\n   # Get table structure\n   {:ok, structure} = Pillar.query(conn, \"DESCRIBE TABLE your_table\")\n   ```\n\n3. **Handle NULL values appropriately:**\n   ClickHouse has strict NULL handling. Use default values where appropriate.\n\n4. **Break into smaller batches:**\n   If inserting large amounts of data, break it into smaller batches:\n   ```elixir\n   Enum.chunk_every(records, 1000)\n   |> Enum.each(fn batch ->\n     Pillar.insert_to_table(conn, \"your_table\", batch)\n   end)\n   ```","ref":"troubleshooting.html#bulk-insert-failures"},{"type":"extras","title":"Async Insert Issues - Troubleshooting","doc":"**Problem:** Async inserts silently fail without error reporting.\n\n**Solutions:**\n\n1. **Implement logging:**\n   While async operations don't return errors, you can log them in your application:\n   ```elixir\n   # Start process monitoring ClickHouse logs\n   def monitor_clickhouse_errors do\n     # Poll error log table periodically\n     schedule_check()\n     \n     # ...\n   end\n   ```\n\n2. **Use buffered inserts with error callbacks:**\n   ```elixir\n   defmodule MyApp.InsertBuffer do\n     use Pillar.BulkInsertBuffer,\n       pool: MyApp.ClickHouse,\n       table_name: \"events\",\n       on_errors: &__MODULE__.handle_errors/2\n       \n     def handle_errors(error, records) do\n       Logger.error(\"Failed insert: #{inspect(error)}\")\n       # Save failed records for retry\n     end\n   end\n   ```","ref":"troubleshooting.html#async-insert-issues"},{"type":"extras","title":"Pool Connection Issues - Troubleshooting","doc":"","ref":"troubleshooting.html#pool-connection-issues"},{"type":"extras","title":"Pool Timeout - Troubleshooting","doc":"**Problem:** Timeout when trying to get a connection from the pool.\n\n**Symptoms:**\n```\n{:error, :timeout}\n```\n\n**Solutions:**\n\n1. **Increase pool size:**\n   ```elixir\n   defmodule MyApp.ClickHouse do\n     use Pillar,\n       connection_strings: [\"http://...\"],\n       pool_size: 30  # Increase from default 10\n   end\n   ```\n\n2. **Increase pool timeout:**\n   ```elixir\n   defmodule MyApp.ClickHouse do\n     use Pillar,\n       connection_strings: [\"http://...\"],\n       pool_timeout: 10_000  # 10 seconds\n   end\n   ```\n\n3. **Check for connection leaks:**\n   Ensure all operations properly return connections to the pool.\n\n4. **Monitor pool utilization:**\n   ```elixir\n   # Check current pool status\n   :poolboy.status(MyApp.ClickHouse.name())\n   ```","ref":"troubleshooting.html#pool-timeout"},{"type":"extras","title":"Schema and Migration Issues - Troubleshooting","doc":"","ref":"troubleshooting.html#schema-and-migration-issues"},{"type":"extras","title":"Migration Failures - Troubleshooting","doc":"**Problem:** ClickHouse migrations fail to apply.\n\n**Solutions:**\n\n1. **Check migration syntax:**\n   Validate SQL statements directly in ClickHouse.\n\n2. **Use transaction-like behavior:**\n   ClickHouse doesn't support transactions, but you can implement rollback functions:\n   ```elixir\n   defmodule Pillar.Migrations.CreateComplexSchema do\n     def up do\n       [\n         \"CREATE TABLE table_1 (...)\",\n         \"CREATE TABLE table_2 (...)\"\n       ]\n     end\n     \n     def down do\n       [\n         \"DROP TABLE table_2\",\n         \"DROP TABLE table_1\"\n       ]\n     end\n   end\n   ```\n\n3. **Inspect migration state:**\n   Query the `pillar_migrations` table to see which migrations were applied:\n   ```elixir\n   {:ok, applied} = Pillar.select(conn, \"SELECT * FROM pillar_migrations\")\n   ```","ref":"troubleshooting.html#migration-failures"},{"type":"extras","title":"Type Conversion Issues - Troubleshooting","doc":"","ref":"troubleshooting.html#type-conversion-issues"},{"type":"extras","title":"DateTime Conversion Problems - Troubleshooting","doc":"**Problem:** DateTime values are not correctly stored or retrieved.\n\n**Solutions:**\n\n1. **Ensure timezone configuration:**\n   ```elixir\n   # In config.exs\n   config :elixir, :time_zone_database, Tzdata.TimeZoneDatabase\n   ```\n\n2. **Use explicit formatting:**\n   ```elixir\n   # When inserting\n   datetime_str = DateTime.utc_now() |> DateTime.to_iso8601()\n   Pillar.query(conn, \"INSERT INTO events VALUES ({timestamp})\", %{timestamp: datetime_str})\n   ```\n\n3. **Customize DateTime conversion:**\n   ```elixir\n   defimpl Pillar.TypeConvert.ToClickHouse, for: DateTime do\n     def convert(datetime) do\n       Calendar.strftime(datetime, \"%Y-%m-%d %H:%M:%S\")\n     end\n   end\n   ```","ref":"troubleshooting.html#datetime-conversion-problems"},{"type":"extras","title":"JSON Data Issues - Troubleshooting","doc":"**Problem:** Complex JSON structures don't serialize correctly.\n\n**Solutions:**\n\n1. **Pre-serialize JSON:**\n   ```elixir\n   json_string = Jason.encode!(complex_map)\n   Pillar.query(conn, \"INSERT INTO logs VALUES ({data})\", %{data: json_string})\n   ```\n\n2. **Use ClickHouse JSON functions:**\n   ```elixir\n   Pillar.select(conn, \"SELECT JSONExtractString(data, 'key') FROM logs\")\n   ```","ref":"troubleshooting.html#json-data-issues"},{"type":"extras","title":"Performance Optimization - Troubleshooting","doc":"","ref":"troubleshooting.html#performance-optimization"},{"type":"extras","title":"Query Performance - Troubleshooting","doc":"1. **Use proper table engines:**\n   - MergeTree family for most analytical workloads\n   - ReplacingMergeTree for data that needs to be updated\n   - SummingMergeTree for pre-aggregated data\n\n2. **Optimize ORDER BY clause:**\n   - Include columns frequently used in WHERE clauses\n   - Put high-cardinality columns first\n\n3. **Use materialized views:**\n   ```sql\n   CREATE MATERIALIZED VIEW user_stats\n   ENGINE = SummingMergeTree()\n   ORDER BY (date, user_id)\n   AS SELECT\n     toDate(timestamp) AS date,\n     user_id,\n     count() AS event_count\n   FROM events\n   GROUP BY date, user_id\n   ```\n\n4. **Add query hints:**\n   ```elixir\n   Pillar.query(conn, \"\"\"\n   SELECT * FROM large_table\n   WHERE date = {date}\n   SETTINGS max_threads = 8, max_memory_usage = 20000000000\n   \"\"\", %{date: Date.utc_today()})\n   ```","ref":"troubleshooting.html#query-performance"},{"type":"extras","title":"Insertion Performance - Troubleshooting","doc":"1. **Batch inserts appropriately:**\n   - Too small: Network overhead\n   - Too large: Memory pressure\n   - Typical sweet spot: 1,000-10,000 records per batch\n\n2. **Use asynchronous inserts for non-critical data:**\n   ```elixir\n   MyApp.ClickHouse.async_insert_to_table(\"logs\", records)\n   ```\n\n3. **Optimize for bulk loads:**\n   ```elixir\n   Pillar.query(conn, \"\"\"\n   INSERT INTO table \n   SELECT * FROM input('format CSV')\n   \"\"\", %{}, %{input_format_allow_errors_num: 10})\n   ```\n\n4. **Consider data locality:**\n   For distributed ClickHouse clusters, insert to the appropriate shard.","ref":"troubleshooting.html#insertion-performance"},{"type":"extras","title":"Connection Optimization - Troubleshooting","doc":"1. **Pool sizing formula:**\n   ```\n   pool_size = min(\n     max_expected_concurrent_queries,\n     (available_memory / avg_query_memory_usage)\n   )\n   ```\n\n2. **Reuse connections:**\n   Avoid creating new connections for every request.\n\n3. **Load balance with multiple servers:**\n   ```elixir\n   defmodule MyApp.ClickHouse do\n     use Pillar,\n       connection_strings: [\n         \"http://clickhouse-1:8123/db\",\n         \"http://clickhouse-2:8123/db\",\n         \"http://clickhouse-3:8123/db\"\n       ]\n   end\n   ```\n\n4. **Use HTTP keep-alive:**\n   Pillar's Tesla Mint adapter uses HTTP keep-alive by default, reducing connection overhead.","ref":"troubleshooting.html#connection-optimization"},{"type":"extras","title":"Debugging Tips - Troubleshooting","doc":"","ref":"troubleshooting.html#debugging-tips"},{"type":"extras","title":"Enable Verbose Logging - Troubleshooting","doc":"```elixir\n# In config/config.exs\nconfig :logger, level: :debug\n\n# In your application\nLogger.configure(level: :debug)\n```","ref":"troubleshooting.html#enable-verbose-logging"},{"type":"extras","title":"Inspect Queries with EXPLAIN - Troubleshooting","doc":"```elixir\nPillar.query(conn, \"EXPLAIN SELECT * FROM large_table WHERE date = {date}\", %{date: \"2023-01-01\"})\n```","ref":"troubleshooting.html#inspect-queries-with-explain"},{"type":"extras","title":"Profile Queries - Troubleshooting","doc":"```elixir\nPillar.query(conn, \"EXPLAIN ANALYZE SELECT * FROM large_table WHERE date = {date}\", %{date: \"2023-01-01\"})\n```","ref":"troubleshooting.html#profile-queries"},{"type":"extras","title":"Monitor System Tables - Troubleshooting","doc":"```elixir\n# Check currently running queries\nPillar.select(conn, \"SELECT * FROM system.processes\")\n\n# Check query log for slow queries\nPillar.select(conn, \"\"\"\n  SELECT \n    query_duration_ms,\n    query \n  FROM system.query_log \n  WHERE type = 'QueryFinish' \n  ORDER BY query_duration_ms DESC \n  LIMIT 10\n\"\"\")\n```","ref":"troubleshooting.html#monitor-system-tables"},{"type":"extras","title":"Server-Side Logging - Troubleshooting","doc":"Increase ClickHouse server log verbosity if needed:\n\n```xml\n<!-- In config.xml -->\n \n     trace \n \n```","ref":"troubleshooting.html#server-side-logging"},{"type":"extras","title":"Common Error Codes - Troubleshooting","doc":"| Code | Description | Common Causes |\n|------|-------------|--------------|\n| 53   | Cannot parse input | Data type mismatch, invalid format |\n| 60   | Invalid query | Syntax error, invalid table name |\n| 62   | Syntax error | Malformed SQL |\n| 81   | Database not found | Wrong database name, incorrect URL |\n| 149  | Tables differ in structure | Schema mismatch in INSERT |\n| 192  | Unknown table | Table doesn't exist |\n| 241  | Memory limit exceeded | Query requires too much memory |\n| 252  | Required server restart | DDL operation needs server restart |","ref":"troubleshooting.html#common-error-codes"},{"type":"extras","title":"Additional Resources - Troubleshooting","doc":"- [ClickHouse Documentation](https://clickhouse.tech/docs/en/)\n- [ClickHouse SQL Reference](https://clickhouse.tech/docs/en/sql-reference/)\n- [System Tables Reference](https://clickhouse.tech/docs/en/operations/system-tables/)\n- [Troubleshooting Performance Issues](https://clickhouse.tech/docs/en/operations/troubleshooting/)","ref":"troubleshooting.html#additional-resources"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.1"}}